/*

ct - a circular tetris-like game
Copyright (C) 2000 Douglas Partridge

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
*/

#include <gnome.h> 
#include <math.h>

#define VERSION "0.11"

#define RINGS 23
#define COLUMNS 13

#define WARP 50

#define MAX_RADIUS 300
#define MIN_RADIUS 50


#define NEXT_PIECE_SIZE 40

enum game_state
{
	GAME_END,
	GAME_PAUSED,
	GAME_RUNNING,
} current_state = GAME_END;

int level;
int speed;
int line_groups_on_level;



gint key_pressed(GtkWidget* widget, GdkEventKey* key);
gint piece_fall_timeout(gpointer data);
void check_rings(void);
void show_score(char*);

GtkWidget* app;
GtkWidget* next_piece_window;
GtkWidget* score_statistics_window;
GtkWidget* piece_statistics_window;

GtkWidget* canvas;
GtkWidget* sbar;
GtkWidget* score_stats;

gint game_new_cb(gpointer);
gint game_pause_cb(gpointer);
gint game_scores_cb(gpointer);
gint game_quit_cb(gpointer);

gint view_next_piece_cb(gpointer);
gint view_piece_statistics_cb(gpointer);
gint view_score_statistics_cb(gpointer);

gint help_about_cb(gpointer);

void display_next_piece(GnomeCanvasItem**, int);
void create_piece_previewer(GnomeCanvasItem***, GtkWidget**, int size);

void display_piece_statistics();



GnomeUIInfo gamemenu[] =
{
	GNOMEUIINFO_MENU_NEW_GAME_ITEM(game_new_cb, NULL),
	GNOMEUIINFO_MENU_PAUSE_GAME_ITEM(game_pause_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_SCORES_ITEM(game_scores_cb, NULL),
	GNOMEUIINFO_MENU_EXIT_ITEM(game_quit_cb, NULL),
	GNOMEUIINFO_END
};

GnomeUIInfo viewmenu[] =
{
	GNOMEUIINFO_TOGGLEITEM("Next piece", "", view_next_piece_cb, NULL),
	GNOMEUIINFO_TOGGLEITEM("Piece statistics", "", view_piece_statistics_cb, NULL),
	GNOMEUIINFO_TOGGLEITEM("Score statistics", "", view_score_statistics_cb, NULL),
	GNOMEUIINFO_END
};

GnomeUIInfo helpmenu[] =
{
	GNOMEUIINFO_MENU_ABOUT_ITEM(help_about_cb, NULL),
	GNOMEUIINFO_END
};

GnomeUIInfo mainmenu[] = {
  	GNOMEUIINFO_MENU_GAME_TREE(gamemenu),
	GNOMEUIINFO_MENU_VIEW_TREE(viewmenu),
	GNOMEUIINFO_MENU_HELP_TREE(helpmenu),
	GNOMEUIINFO_END
};

typedef struct _segment segment;
typedef struct _piece piece;

struct _segment
{
	gboolean occupied;
	GnomeCanvasItem* p;
	char const* colour;
};

struct _piece
{
	int column;
	int ring;

	int sections[4][2];
	int type;
	int next_type;

	char const* colour;
};

piece current_piece;

int score[5];
int total_score;

int piece_count[8];
GtkWidget* piece_stats_labels[8];

const int points[] = {1, 50, 150, 300, 600};
const char* point_names[] = {"drop points", "singles", "doubles", "triples", "quads"};
segment field[RINGS][COLUMNS];

GnomeCanvasItem** next_piece;


const int pieces[7][4][2] = {{{ 0, 0}, { 0,  1}, { 1,  0}, {1, 1}},
			     {{ 0, 0}, {-1,  0}, { 0,  1}, {1, 1}},
			     {{ 0, 0}, { 0,  1}, {-1,  1}, {1, 0}},
			     {{ 0, 0}, { 0, -1}, { 0,  1}, {1, 0}},
			     {{ 0, 0}, { 0, -1}, { 1, -1}, {0, 1}},
			     {{ 0, 0}, { 0, -1}, { 0,  1}, {1, 1}},
			     {{ 0, 0}, { 0, -1}, { 0,  1}, {0, 2}}};

const char* colours[7] = {"red", "green", "blue", "yellow", "orange", "cyan", "purple"};



void
randomise_piece(piece* ci)
{
	int c;

	ci->column = (int)((double)COLUMNS * rand()/(RAND_MAX + 1.0));

	ci->type = ci->next_type;
	piece_count[ci->type]++;
	piece_count[7]++;
	ci->colour = colours[ci->type];
	memcpy(ci->sections, pieces[ci->type], 4 * 2 * sizeof(int));

	c = (int) (7.0 * rand()/(RAND_MAX + 1.0));
	ci->next_type = c;
	display_next_piece(next_piece, ci->next_type);
	display_piece_statistics();
}



GnomeCanvasPoints*
create_points(int ring, int column, int res)
{
	GnomeCanvasPoints* points;
	ArtPoint p1, p2;
	double rotor[6];
	int i;

	// it would be possible to calculate some of these values only once
	double column_angular_size = 360 / (double)COLUMNS;

	double rpos = pow((ring / (double)RINGS), 1.2);
	double nrpos = pow(((ring + 1) / (double)RINGS), 1.2);
	
	double min_ring_radius =  rpos * (MAX_RADIUS - MIN_RADIUS) + MIN_RADIUS;
	double max_ring_radius = nrpos * (MAX_RADIUS-MIN_RADIUS)+MIN_RADIUS;

	double column_start_angle = (double)column * column_angular_size + (rpos * WARP);
	double column_end_angle = (double)(column + 1) * column_angular_size + (nrpos * WARP);

	points = gnome_canvas_points_new(res * 2);


	// it would be possible to do this with one loop insted of two
	p1.x = min_ring_radius;
	p1.y = 0.0;
	for(i = 0; i < res; i++)
	{
		art_affine_rotate(rotor, (i / (double)(res - 1)) * column_angular_size + column_start_angle);
		art_affine_point(&p2, &p1, rotor);

		points->coords[2 * i] = p2.x;
		points->coords[2 * i + 1] = p2.y;
	}


	p1.x = max_ring_radius;
	p1.y = 0.0;
	for(i = 0; i < res; i++)
	{
		art_affine_rotate(rotor, column_end_angle - (i / (double)(res - 1)) * column_angular_size);
		art_affine_point(&p2, &p1, rotor);

		points->coords[2 * i + 2 * res] = p2.x;
		points->coords[2 * i + 2 * res + 1] = p2.y;
	}
	
	return points;
}


void
create_field()
{
	static gboolean first = TRUE;
	int r, c;
	GnomeCanvasPoints* points;
	GnomeCanvasItem* w;


	if (first)
	{
		gnome_canvas_item_new(gnome_canvas_root(GNOME_CANVAS(canvas)),
				      gnome_canvas_rect_get_type(),
				      "x1", (gdouble)-MAX_RADIUS - 1,
				      "y1", (gdouble)-MAX_RADIUS - 1,
				      "x2", (gdouble) MAX_RADIUS + 1,
				      "y2", (gdouble) MAX_RADIUS + 1,
				      "fill_color", "black",
				      NULL);

		gnome_canvas_item_new(gnome_canvas_root(GNOME_CANVAS(canvas)),
				      gnome_canvas_ellipse_get_type(),
				      "x1", (gdouble)-MAX_RADIUS,
				      "y1", (gdouble)-MAX_RADIUS,
				      "x2", (gdouble) MAX_RADIUS,
				      "y2", (gdouble) MAX_RADIUS,
				      "fill_color", "grey",
				      NULL);

		gnome_canvas_item_new(gnome_canvas_root(GNOME_CANVAS(canvas)),
				      gnome_canvas_ellipse_get_type(),
				      "x1", (gdouble)-MIN_RADIUS,
				      "y1", (gdouble)-MIN_RADIUS,
				      "x2", (gdouble) MIN_RADIUS,
				      "y2", (gdouble) MIN_RADIUS,
				      "fill_color", "black",
				      NULL);
	}


	for(r = 0; r < RINGS; r++)
	{
		for(c = 0; c < COLUMNS; c++)
		{
			if (first)
			{
				points = create_points(r, c, 4);
				w = gnome_canvas_item_new(gnome_canvas_root(GNOME_CANVAS(canvas)),
						  	gnome_canvas_polygon_get_type(),
							  "GnomeCanvasPolygon::points", points,
							  "GnomeCanvasPolygon::fill_color", "red",
							  "GnomeCanvasPolygon::outline_color", "black",
							  "GnomeCanvasPolygon::width_pixels", 2,
							  NULL);
				gnome_canvas_points_free(points);
				field[r][c].p = w;
			}



			gnome_canvas_item_hide(field[r][c].p);
			field[r][c].occupied = FALSE;
		}
	}

	first = FALSE;
}


static void
game_maybe_quit(GtkWidget *widget, int button)
{
 	if (button == 0)
	{
 	 	gtk_widget_destroy(app);
 	 	gtk_main_quit();
 	}

}


gint
game_new_cb(gpointer data)
{
	int i;


	gtk_widget_set_sensitive(gamemenu[1].widget, TRUE);

	total_score = 0;
	for(i = 0; i < 5; i++)
	{
		score[i] = 0;
	}
	show_score("");

	for(i = 0; i < 8; i++)
	{
		piece_count[i] = 0;
	}

	current_piece.ring = RINGS - 1;
	current_piece.next_type = (int) (7.0 * rand()/(RAND_MAX + 1.0));
	randomise_piece(&current_piece);

	display_piece_statistics();

	level = 1;
	line_groups_on_level = 0;

	create_field();

	if(current_state == GAME_END)
	{
		speed = 1000;
		gtk_timeout_add(speed, piece_fall_timeout, NULL);
	}
	current_state = GAME_RUNNING;

	return TRUE;
}


gint
game_pause_cb(gpointer data)
{
	switch (current_state)
	{
		case GAME_PAUSED:
			current_state = GAME_RUNNING;
			gtk_timeout_add(speed, piece_fall_timeout, NULL);
			break;

		case GAME_RUNNING:
			current_state = GAME_PAUSED;
			break;

		default:
			break;
	}

	return TRUE;
}

gint
game_scores_cb(gpointer data)
{
	gnome_scores_display("ct", "ct", NULL, 0);

	return TRUE;
}

gint
game_quit_cb(gpointer data)
{
	GtkWidget *box;

	box = gnome_message_box_new (_("Do you really want to quit?"),
	                             GNOME_MESSAGE_BOX_QUESTION,
	                             GNOME_STOCK_BUTTON_YES,
	                             GNOME_STOCK_BUTTON_NO,
	                             NULL);
	gnome_dialog_set_parent (GNOME_DIALOG(box), GTK_WINDOW(app));
	gnome_dialog_set_default (GNOME_DIALOG (box), 0);
	gtk_window_set_modal (GTK_WINDOW (box), TRUE);  
	gtk_signal_connect (GTK_OBJECT (box), "clicked",
	                   (GtkSignalFunc)game_maybe_quit, NULL);
	gtk_widget_show (box);

 	return TRUE;
}


gint
view_next_piece_cb(gpointer data)
{
	if(GTK_CHECK_MENU_ITEM(viewmenu[0].widget)->active)
	{
		gtk_widget_show(next_piece_window);
	}
	else
	{
		gtk_widget_hide(next_piece_window);
	}

	return TRUE;
}


gint
view_piece_statistics_cb(gpointer data)
{
	if(GTK_CHECK_MENU_ITEM(viewmenu[1].widget)->active)
	{
		gtk_widget_show(piece_statistics_window);
	}
	else
	{
		gtk_widget_hide(piece_statistics_window);
	}


	return TRUE;
}

gint
view_score_statistics_cb(gpointer data)
{
	if(GTK_CHECK_MENU_ITEM(viewmenu[2].widget)->active)
	{
		gtk_widget_show(score_statistics_window);
	}
	else
	{
		gtk_widget_hide(score_statistics_window);
	}

	return TRUE;
}


gint
help_about_cb(gpointer data)
{
	const gchar* authors[] = {"Douglas Partridge <dpartrid@lawson.comp.utas.edu.au>", NULL};


	gtk_widget_show(gnome_about_new("ct", VERSION, 
				        "Copyright 2000 Douglas Partridge",
					authors,
					"nothing",
					NULL));

	return TRUE;
}



GtkWidget*
create_main_window(void)
{
	GtkWidget* a;

	canvas = gnome_canvas_new();
	gnome_canvas_set_scroll_region(GNOME_CANVAS(canvas), -200, -200, 200, 200);
	gtk_widget_set_usize(GTK_WIDGET(canvas), 2 * MAX_RADIUS + 5, 2 * MAX_RADIUS + 5);
	create_field(GNOME_CANVAS(canvas));


	sbar = gnome_appbar_new(FALSE, TRUE, GNOME_PREFERENCES_USER);


	a = gnome_app_new("ct", "ct");
	gnome_app_set_contents(GNOME_APP(a), canvas);
	gnome_app_set_statusbar(GNOME_APP(a), sbar);

	gtk_widget_set(a,
		       "GtkObject::signal::delete_event", game_quit_cb, NULL,
		       "GtkObject::signal::key_press_event", key_pressed, NULL,
		       "GtkWindow::allow_grow", FALSE,
		       "GtkWindow::allow_shrink", FALSE,
		       NULL);

	gnome_app_create_menus(GNOME_APP(a), mainmenu);
	gnome_app_install_menu_hints(GNOME_APP(a), mainmenu);

	gtk_widget_set_sensitive(gamemenu[1].widget, FALSE);


	gtk_widget_show_all(a);

	return a;
}


gboolean
false(gpointer data)
{
	return FALSE;
}

gboolean
true(gpointer data)
{
	return TRUE;
}

gboolean
next_piece_window_delete_cb(gpointer data)
{
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(viewmenu[0].widget), FALSE);

	return TRUE;
}


gboolean
piece_statistics_window_delete_cb(gpointer data)
{
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(viewmenu[1].widget), FALSE);

	return TRUE;
}


gboolean
score_statistics_window_delete_cb(gpointer data)
{
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(viewmenu[2].widget), FALSE);

	return TRUE;
}


void
create_piece_previewer(GnomeCanvasItem*** p, GtkWidget** c, int size)
{
	int i, j;

	*p = g_malloc(4 * 3 * sizeof(GnomeCanvasItem*));

	*c = gnome_canvas_new();
	gnome_canvas_set_scroll_region(GNOME_CANVAS(*c), 0, 0, 4 * size, 3 * size);
	gtk_widget_set_usize(GTK_WIDGET(*c), 4 * size, 3 * size);

	for(i = 0; i < 4; i++)
	{
		for(j = 0; j < 3; j++)
		{
			(*p)[3 * i + j] =
			gnome_canvas_item_new(gnome_canvas_root(GNOME_CANVAS(*c)),
					      gnome_canvas_rect_get_type(),
					      "x1", (gdouble)(i * size),
					      "y1", (gdouble)(j * size),
					      "x2", (gdouble)(i * size) + size,
					      "y2", (gdouble)(j * size) + size,
					      "fill_color", "red",
					      "outline_color", "black",
					      "width_pixels", 2,
					      NULL);
			gnome_canvas_item_hide((*p)[3 * i + j]);
		}
	}

	gtk_widget_show_all(*c);
}

GtkWidget*
create_next_piece_window()
{
	GtkWidget* w;
	GtkWidget* c;

	w = gtk_widget_new(gtk_window_get_type(),
			   "title", "Next piece",
			   "GtkObject::signal::delete_event", next_piece_window_delete_cb,
			   "GtkWindow::allow_grow", FALSE,
			   "GtkWindow::allow_shrink", FALSE,
			   NULL);
		       

	create_piece_previewer(&next_piece, &c, NEXT_PIECE_SIZE);
	gtk_container_add(GTK_CONTAINER(w), c);


	gtk_widget_show_all(c);

	return w;
}


GtkWidget*
create_score_statistics_window()
{
	int i;
	GtkWidget* w;

	const char* titles[] = {"type", "number", "value", "total"};
	char* tmp[4] = {"", "", "", "0"};


	w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(w), "Score statistics");
	gtk_signal_connect(GTK_OBJECT(w),
			   "delete_event", GTK_SIGNAL_FUNC(score_statistics_window_delete_cb),
			   NULL);


	score_stats = gtk_clist_new_with_titles(4, (gchar**)titles);

	gtk_clist_set_column_width(GTK_CLIST(score_stats), 0, 65);
	for(i = 1; i < 4; i++)
	{
		gtk_clist_set_column_justification(GTK_CLIST(score_stats), i, GTK_JUSTIFY_RIGHT);
		gtk_clist_set_column_width(GTK_CLIST(score_stats), i, 65);
	}

	tmp[1] = g_strdup("0");
	for(i = 0; i < 5; i++)
	{
		char buff[100];

		tmp[0] = g_strdup(point_names[i]);
		sprintf(buff, "%d", points[i]);
		tmp[2] = g_strdup(buff);
		gtk_clist_append(GTK_CLIST(score_stats), (gchar**)tmp);
		g_free(tmp[0]);
		g_free(tmp[2]);
	}
	tmp[0] = g_strdup("");
	g_free(tmp[1]);
	tmp[1] = g_strdup("");
	tmp[2] = g_strdup("grand total");
	gtk_clist_append(GTK_CLIST(score_stats), (gchar**)tmp);
	g_free(tmp[0]);
	g_free(tmp[2]);

	gtk_container_add(GTK_CONTAINER(w), score_stats);
	gtk_widget_show(score_stats);

	return w;
}


GtkWidget*
create_piece_statistics_window()
{

	GtkWidget* w;
	GtkWidget* t;
	GtkWidget* l;
	int i;
	char buff[100];

	w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(w), "Piece statistics");
	gtk_signal_connect(GTK_OBJECT(w),
			   "delete_event", GTK_SIGNAL_FUNC(piece_statistics_window_delete_cb),
			   NULL);

	t = gtk_table_new(8, 2, TRUE);

	for(i = 0; i < 7; i++)
	{
		GnomeCanvasItem** p;
		GtkWidget* c;

		create_piece_previewer(&p, &c, 20);
		display_next_piece(p, i);
		g_free(p);

		gtk_table_attach_defaults(GTK_TABLE(t), c, 0, 1, i, i+1);

		sprintf(buff, "%d", piece_count[i]);
		l = gtk_label_new(buff);
		gtk_table_attach_defaults(GTK_TABLE(t), l, 1, 2, i, i+1);

		piece_stats_labels[i] = l;
	}

	l = gtk_label_new("total");
	gtk_table_attach_defaults(GTK_TABLE(t), l, 0, 1, 7, 8);

	sprintf(buff, "%d", piece_count[i]);
	l = gtk_label_new(buff);
	gtk_table_attach_defaults(GTK_TABLE(t), l, 1, 2, 7, 8);

	piece_stats_labels[7] = l;


	gtk_container_add(GTK_CONTAINER(w), t);

	gtk_widget_show_all(t);
	return w;
}





void
display_next_piece(GnomeCanvasItem** p, int type)
{
	int i, j;
	//g_print("hello\n");
	for(i = 0; i < 4; i++)
	{
		for(j = 0; j < 3; j++)
		{
			gnome_canvas_item_hide(p[i * 3 + j]);
		}
	}

	for(i = 0; i < 4; i++)
	{
		int x = 3 - (pieces[type][i][1] + 1);
		int y = pieces[type][i][0] + 1;

		//g_print("%d, %d\n", x, y);
		gnome_canvas_item_set(p[3 * x + y],
				      "fill_color", colours[type],
				      NULL);
		gnome_canvas_item_show(p[3 * x + y]);
	}
}


void
display_piece_statistics()
{
	int i;

	for(i = 0; i < 8; i++)
	{
		char buff[100];
		sprintf(buff, "%d", piece_count[i]);
		gtk_label_set_text(GTK_LABEL(piece_stats_labels[i]), buff);
	}
}

void
clamp_column(int* col)
{
	if (*col < 0) *col += COLUMNS;
		else if (*col >= COLUMNS) *col -= COLUMNS;
}

void
piece_clear()
{
	int i;
	int tmp_rng, tmp_col;
	for(i = 0; i < 4; i++)
	{
		tmp_rng = current_piece.ring + current_piece.sections[i][0];
		tmp_col = current_piece.column + current_piece.sections[i][1];

		if (tmp_rng >= RINGS) continue;
		clamp_column(&tmp_col);

		gnome_canvas_item_hide(field[tmp_rng][tmp_col].p);
	}
}

void
piece_set()
{
	int i;
	int tmp_rng, tmp_col;
	for(i = 0; i < 4; i++)
	{
		tmp_rng = current_piece.ring + current_piece.sections[i][0];
		tmp_col = current_piece.column + current_piece.sections[i][1];

		if (tmp_rng >= RINGS) continue;
		clamp_column(&tmp_col);

		gnome_canvas_item_set(field[tmp_rng][tmp_col].p,
				      "fill_color", current_piece.colour,
				      NULL);
		gnome_canvas_item_show(field[tmp_rng][tmp_col].p);
		field[tmp_rng][tmp_col].colour = current_piece.colour;
	}
}

void
piece_move(int c)
{
	int new_col;
	int tmp_col, tmp_rng;
	int i;

	g_assert(c == 1 || c == -1);


	new_col = current_piece.column + c;

	clamp_column(&new_col);

	for(i = 0; i < 4; i++)
	{
		tmp_rng = current_piece.ring + current_piece.sections[i][0];
		tmp_col = new_col + current_piece.sections[i][1];

		if (tmp_rng >= RINGS) continue;

		clamp_column(&tmp_col);

		if (field[tmp_rng][tmp_col].occupied)
			return;
	}

	piece_clear();
	current_piece.column = new_col;
	piece_set();
	
	return;
}

void
piece_fall()
{
	int new_rng;
	int tmp_rng, tmp_col;
	int i;
	gboolean is_new_piece;
	is_new_piece = FALSE;

	new_rng = current_piece.ring - 1;

	for (i = 0; i < 4; i++)
	{
		tmp_rng = new_rng + current_piece.sections[i][0];
		tmp_col = current_piece.column + current_piece.sections[i][1];

		if (tmp_rng >= RINGS) continue;
		clamp_column(&tmp_col);

		if (tmp_rng < 0 || field[tmp_rng][tmp_col].occupied)
		{
			// the piece has come to rest
			int j;
			for(j = 0; j < 4; j++)
			{
				int tmp_rng, tmp_col;

				// mark it as occupied
				tmp_rng = current_piece.ring + current_piece.sections[j][0];
				tmp_col = current_piece.column + current_piece.sections[j][1];

				if (tmp_rng >= RINGS) continue;
				//g_assert(tmp_rng > 0);

				clamp_column(&tmp_col);

				field[tmp_rng][tmp_col].occupied = TRUE;
			}

			// now create a new one
			current_piece.ring = RINGS - 1;

			randomise_piece(&current_piece);

			check_rings();

			is_new_piece = TRUE;
			break;
			//return;
		}
	}

	if (is_new_piece)
	{
		int i;
		for(i = 0; i < 4; i++)
		{
			tmp_rng = current_piece.ring + current_piece.sections[i][0];
			tmp_col = current_piece.column + current_piece.sections[i][1];

			if (tmp_rng >= RINGS) continue;
			clamp_column(&tmp_col);

			if (field[tmp_rng][tmp_col].occupied)
			{

				gtk_widget_set_sensitive(gamemenu[1].widget, FALSE);
				show_score("game over");

				if(current_state == GAME_RUNNING)
				{
					int pos;

					pos = gnome_score_log((gfloat)total_score, NULL, TRUE);
					gnome_scores_display("ct", "ct", NULL, pos);
				}

				current_state = GAME_END;
				//g_print("game end\n");
			}
		}
	}
	else
	{
		piece_clear();

		current_piece.ring = new_rng;
	}


	piece_set();

	return;
}


void
piece_rotate(int direction)
{
	int new_piece[4][2];
	int i;

	g_assert(direction == 1 || direction == -1);

	for(i = 0; i < 4; i++)
	{
		new_piece[i][0] = -direction * current_piece.sections[i][1];
		new_piece[i][1] = direction * current_piece.sections[i][0];

		if (field[current_piece.ring+new_piece[i][0]][current_piece.column+new_piece[i][1]].occupied)
		{
			return;
		}
	}


	piece_clear();
	memcpy(current_piece.sections, new_piece, 4 * 2 * sizeof(int));
	piece_set();

}


void
check_rings(void)
{
	int i, j;
	gboolean ring_complete;
	int num_rings = 0;

	for(i = RINGS - 1; i >= 0; i--)
	{
		for(j = 0; j < COLUMNS; j++)
		{
			ring_complete = TRUE;
			if (!field[i][j].occupied)
			{
				ring_complete = FALSE;
				break;
			}
		}

		if (ring_complete)
		{
			int r, c;

			num_rings++;

			for(r = i; r < RINGS - 1; r++)
			{
				for(c = 0; c < COLUMNS; c++)
				{
					field[r][c].occupied = field[r + 1][c].occupied;
					field[r][c].colour = field[r + 1][c].colour;
					gnome_canvas_item_set(field[r][c].p,
							      "fill_color", field[r+1][c].colour,
							      NULL);
					if (field[r][c].occupied)
					{
						gnome_canvas_item_show(GNOME_CANVAS_ITEM(field[r][c].p));
					}
					else
					{
						gnome_canvas_item_hide(GNOME_CANVAS_ITEM(field[r][c].p));
					}
				}
			}
		}
	}

	if (num_rings != 0)
	{
		line_groups_on_level++;

		score[num_rings]++;
		total_score += points[num_rings];
		show_score("");
	}
}


gint
piece_fall_timeout(gpointer data)
{
	if (current_state == GAME_RUNNING)
	{
		piece_fall();
		if (line_groups_on_level > 2)
		{
			speed *= 0.75;

			line_groups_on_level = 0;
			level++;

			show_score("");
			//g_print("changing timeout to %d\n", speed);

			gtk_timeout_add(speed, piece_fall_timeout, NULL);
			return FALSE;
		}
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}


gint
key_pressed(GtkWidget* widget, GdkEventKey* key)
{
	if (current_state != GAME_RUNNING)
	{
		return TRUE;
	}
	if (key->keyval == GDK_Left)
	{
		piece_move(-1);
	}
	else if (key->keyval == GDK_Right)
	{
		piece_move(1);
	}
	else if (key->keyval == GDK_space)
	{
		score[0]++;
		total_score += points[0];
		show_score("");
		piece_fall();
	}
	else if (key->keyval == GDK_Up)
	{
		piece_rotate(-1);
	}
	else if (key->keyval == GDK_Down)
	{
		piece_rotate(1);
	}

	return TRUE;
}

void
show_score(char* message)
{
	int i;
	char buff[200];

	sprintf(buff, "score : %d\t\t\t"
		      "level : %d\t\t\t"
		      "ring groups : %d/3"
		      "\t\t\t%s",
		      total_score, level, line_groups_on_level, message);

	gnome_appbar_set_status(GNOME_APPBAR(sbar), buff);

	for(i = 0; i < 5; i++)
	{
		sprintf(buff, "%d", score[i]);
		gtk_clist_set_text(GTK_CLIST(score_stats), i, 1, buff);

		sprintf(buff, "%d", score[i] * points[i]);
		gtk_clist_set_text(GTK_CLIST(score_stats), i, 3, buff);

	}
	sprintf(buff, "%d", total_score);
	gtk_clist_set_text(GTK_CLIST(score_stats), 5, 3, buff);
	
}

int main(int argc, char** argv)
{
	g_print("%d\n", gnome_score_init("ct"));
	srand(time(NULL));

	gnome_init("ct", VERSION, argc, argv);

	app = create_main_window();
	next_piece_window = create_next_piece_window();
	score_statistics_window = create_score_statistics_window();
	piece_statistics_window = create_piece_statistics_window();


	gtk_main();

	return 0;
}

